import sys
from dataclasses import dataclass
from typing import Annotated, Literal, Optional, Union

import pytest
from serpyco_rs import Serializer
from serpyco_rs.metadata import Discriminator


@dataclass
class Foo:
    val: int
    type: Literal["foo"] = "foo"


@dataclass
class Bar:
    type: Literal["bar"]
    val: str


@dataclass
class Buz:
    type: int
    val: str


@dataclass
class Buzz:
    val: str


@dataclass
class Base:
    childs: list[Annotated[Union[Foo, Bar], Discriminator("type")]]


def test_tagged_union():
    serializer = Serializer(Base)
    val = Base(
        childs=[
            Foo(val=1),
            Bar(type="bar", val="12"),
        ]
    )
    raw = {"childs": [{"type": "foo", "val": 1}, {"type": "bar", "val": "12"}]}
    assert serializer.dump(val) == raw
    assert serializer.load(raw) == val


def test_tagged_union__invalid_discriminator_type():
    @dataclass
    class Inner:
        field: Annotated[Union[Foo, Bar, Buz], Discriminator("type")]

    with pytest.raises(RuntimeError) as exc_info:
        Serializer(Inner)

    assert exc_info.type is RuntimeError
    assert exc_info.value.args[0] == (
        "Type <class 'tests.test_union.Buz'> has invalid discriminator field \"type\" with type \"<class 'int'>\". "
        "Discriminator supports only Literal[<str>]."
    )


def test_tagged_union__union_arg_has_no_discriminator_field():
    @dataclass
    class Inner:
        field: Annotated[Union[Foo, Bar, Buzz], Discriminator("type")]

    with pytest.raises(RuntimeError) as exc_info:
        Serializer(Inner)

    assert exc_info.type is RuntimeError
    assert exc_info.value.args[0] == ("Type <class 'tests.test_union.Buzz'> does not have discriminator field \"type\"")


def test_tagged_union__unsupported_types():
    @dataclass
    class Inner:
        field: Annotated[Union[int, str], Discriminator("type")]

    with pytest.raises(RuntimeError) as exc_info:
        Serializer(Inner)

    assert exc_info.type is RuntimeError
    assert exc_info.value.args[0] == (
        "Unions supported only for dataclasses or attrs. Provided: typing.Union[<class 'int'>,<class 'str'>]"
    )


@dataclass
class A:
    type: Literal["A"]
    params: int
    children: Optional[list["ComponentsT"]] = None


@dataclass
class B:
    type: Literal["B"]
    params: str
    children: Optional[list["ComponentsT"]] = None


ComponentsT = Annotated[Union[A, B], Discriminator("type")]


@pytest.mark.skipif(sys.version_info < (3, 11), reason="requires python3.11 or higher")
def test_tagged_union__with_forward_refs():
    serializer = Serializer(ComponentsT, omit_none=True)
    data: ComponentsT = A(
        type="A",
        params=123,
        children=[
            A(
                type="A",
                params=1234,
            ),
            B(
                type="B",
                params="foo",
                children=[
                    B(
                        type="B",
                        params="bar",
                    )
                ],
            ),
        ],
    )
    raw_data = {
        "type": "A",
        "params": 123,
        "children": [
            {"params": 1234, "type": "A"},
            {"children": [{"params": "bar", "type": "B"}], "params": "foo", "type": "B"},
        ],
    }
    assert serializer.dump(data) == raw_data
    assert serializer.load(raw_data) == data


def test_tagged_union__camel_case_filed_format():
    @dataclass
    class A:
        discriminator_with_multiple_words: Literal["A"]
        a: int

    @dataclass
    class B:
        discriminator_with_multiple_words: Literal["B"]
        b: str

    raw_obj = [
        {
            "discriminatorWithMultipleWords": "A",
            "a": 128,
        },
        {
            "discriminatorWithMultipleWords": "B",
            "b": "foo",
        },
    ]
    obj = [
        A(
            discriminator_with_multiple_words="A",
            a=128,
        ),
        B(
            discriminator_with_multiple_words="B",
            b="foo",
        ),
    ]

    serializer = Serializer(
        list[Annotated[Union[A, B], Discriminator("discriminator_with_multiple_words")]],
        camelcase_fields=True,
    )

    assert serializer.dump(obj) == raw_obj
    assert serializer.load(raw_obj) == obj



from .compare.libs import serpyco_rs as serializer
from .utils import repeat


def test_dump(benchmark):
    serializer.dump(serializer.test_object)  # warmup
    benchmark.group = "dump"
    benchmark(repeat(lambda: serializer.dump(serializer.test_object), count=100))


def test_load(benchmark):
    test_dict = serializer.dump(serializer.test_object)
    serializer.load(test_dict, validate=False)  # warmup

    benchmark.group = "load"
    benchmark(repeat(lambda: serializer.load(test_dict, validate=False), count=100))


def test_load_validate(benchmark):
    test_dict = serializer.dump(serializer.test_object)
    serializer.load(test_dict, validate=True)  # warmup

    benchmark.group = "load with validate"
    benchmark(repeat(lambda: serializer.load(test_dict, validate=True), count=100))



from dataclasses import dataclass
from typing import Annotated, Optional

import pytest
from serpyco_rs import Serializer
from serpyco_rs.metadata import Alias, OmitNone


def test_annotated_filed_alias():
    @dataclass
    class A:
        foo: Annotated[str, Alias("bar")]

    serializer = Serializer(A)

    obj = A(foo="123")

    expected = {"bar": "123"}
    assert serializer.dump(obj) == expected
    assert serializer.load(expected) == obj


def test_omit_none():
    @dataclass
    class A:
        required_val: Optional[bool]
        optional_val: Optional[bool] = None

    serializer = Serializer(Annotated[A, OmitNone])

    expected = {"required_val": None}
    assert serializer.dump(A(required_val=None)) == expected
    assert serializer.load(expected) == A(required_val=None)


@dataclass
class Bar:
    value: Optional[str] = None


@dataclass
class Foo:
    bar: Optional[Bar] = None


@pytest.mark.parametrize(
    ["omit_none", "foo", "expected"],
    [
        (True, Foo(), {}),
        (True, Foo(Bar()), {"bar": {}}),
        (False, Foo(), {"bar": None}),
        (False, Foo(Bar()), {"bar": {"value": None}}),
    ],
)
def test_propagete__omit_none(omit_none, foo, expected):
    serializer = Serializer(Foo, omit_none=omit_none)
    res = serializer.dump(foo)

    assert res == expected


def test_omit_none_on_dict():
    serializer = Serializer(dict[str, Optional[bool]], omit_none=True)
    assert serializer.dump({"foo": True, "bar": None}) == {"foo": True}



from collections.abc import Callable
from dataclasses import dataclass
from decimal import Decimal
from enum import Enum
from typing import Generic, Optional, TypeVar, Union


@dataclass(frozen=True)
class Min:
    value: Union[int, float, Decimal]


@dataclass(frozen=True)
class Max:
    value: Union[int, float, Decimal]


@dataclass(frozen=True)
class MinLength:
    value: int


@dataclass(frozen=True)
class MaxLength:
    value: int


@dataclass(frozen=True)
class Places:
    value: int


@dataclass(frozen=True)
class Discriminator:
    name: str


@dataclass(frozen=True)
class Alias:
    value: str


class Format(Enum):
    no_format = "no_format"
    camel_case = "camel_case"


@dataclass(frozen=True)
class FiledFormat:
    format: Format


CamelCase: FiledFormat = FiledFormat(Format.camel_case)
NoFormat: FiledFormat = FiledFormat(Format.no_format)


@dataclass(frozen=True)
class NoneFormat:
    omit: bool


KeepNone: NoneFormat = NoneFormat(False)
OmitNone: NoneFormat = NoneFormat(True)


_I = TypeVar("_I")
_O = TypeVar("_O")


@dataclass(frozen=True)
class CustomEncoder(Generic[_I, _O]):
    serialize: Optional[Callable[[_I], _O]] = None
    deserialize: Optional[Callable[[_O], _I]] = None


def serialize_with(func: Callable[[_I], _O]) -> CustomEncoder[_I, _O]:
    return CustomEncoder[_I, _O](serialize=func)


def deserialize_with(func: Callable[[_O], _I]) -> CustomEncoder[_I, _O]:
    return CustomEncoder[_I, _O](deserialize=func)
